from __future__ import annotations

import logging
from typing import Optional, Callable

from rx import operators as ops
from rx.core import ConnectableObservable
from rx.core.abc import Disposable

from sfm2._implementation.commands.command import Command
from sfm2._implementation.internal_utils.wrap_exception import wrap_exception
from sfm2._implementation.serial_port.serial_port import SerialPort
from sfm2.interface.exceptions import CommandResponseTimeoutException, CommandLineTimeoutException

_log = logging.getLogger(__name__)


class WaitForResponse:
    def __init__(self, port: SerialPort, cmd: Command, timeout_s: float):
        self._impl = _Impl(port, lambda l: l.startswith(cmd.name))
        self._cmd = cmd
        self._timeout_s: float = timeout_s

    def _create_timeout(self):
        return CommandResponseTimeoutException(
            command_name=self._cmd.name,
            timeout_s=self._timeout_s)

    @wrap_exception(create_wrapping_exc=_create_timeout)
    def _wait(self) -> str:
        line = self._impl.wait(self._timeout_s)
        return self._cmd.strip_value(line)

    def __enter__(self) -> Callable[[], str]:
        self._impl.connect()
        return self._wait

    def __exit__(self, typ, value, traceback):
        self._impl.dispose()

    def _create_timeout(self):
        return CommandLineTimeoutException(
            timeout_s=self._timeout_s)


class WaitForLine:
    def __init__(self, port: SerialPort, line_matcher: Callable[[str], bool], timeout_s: float):
        self._timeout_s: float = timeout_s
        self._impl = _Impl(port, line_matcher)

    def __enter__(self) -> Callable[[], str]:
        self._impl.connect()
        return self._wait

    def __exit__(self, typ, value, traceback):
        self._impl.dispose()

    def _create_timeout(self):
        return CommandLineTimeoutException(
            timeout_s=self._timeout_s)

    @wrap_exception(create_wrapping_exc=_create_timeout)
    def _wait(self) -> str:
        return self._impl.wait(self._timeout_s)


class _Impl:
    def __init__(self, port: SerialPort, line_matcher: Callable[[str], bool]):
        self._disposable: Optional[Disposable] = None
        self._response: ConnectableObservable[str] = port.incoming_text.pipe(
            ops.filter(line_matcher),
            ops.take(1),
            ops.map(str.strip),
            ops.replay(buffer_size=1)
        )

    def connect(self) -> None:
        self._disposable = self._response.connect()

    def wait(self, timeout_s: float) -> str:
        result = self._response.pipe(
            ops.timeout(timeout_s)
        ).run()
        return result

    def dispose(self) -> None:
        if self._disposable is not None:
            self._disposable.dispose()
